// Sundog Resurection : Sundog Navigator

import java.util.*;
import java.io.*;
import java.awt.event.*;
import java.awt.*;
import java.awt.image.*;
import java.net.*;

/**
 * The proto Sundog Navigator
 */
public class SundogNavigator
{
    /** Used for Double Buffering */
    public static Dimension offDim;
    public static Image     offImage;
    public static Graphics  offGraphics;

    public static Sector currentSector;
    public static StarSystem currentStarSystem;
    public static Planet currentPlanet;

    /** Matrix entry */
    public static double
	v1M11, v1M12, v1M13,
	v1M21, v1M22, v1M23,
	v1M31, v1M32, v1M33;
    public static double
	v2M11, v2M12, v2M13,
	v2M21, v2M22, v2M23,
	v2M31, v2M32, v2M33;
    public static double
	v3M11, v3M12, v3M13,
	v3M21, v3M22, v3M23,
	v3M31, v3M32, v3M33;

    /** Orientation of point of view in mode given by the number*/
    public static double theta1,phi1;
    public static double theta2,phi2;
    public static double theta3,phi3;

    /** This is to remind the initial orientation while turning*/
    public static double theta_0,phi_0;

    public static final double M=50;
    public static final double H=25;
    public static final double angleFactor=0.005;
    public static final double angleFactor3=0.010;
    public static final double scale10=0.02; // Unit : Ly
    public static final double scale20=0.01; // Unit : AU
    public static final double scale30=0.0001; // Unit : km
    public static final int RAD=4;

    public static final double AU=40.0;

    public static boolean box;
    public static boolean names;
    public static boolean hideText;
    public static boolean sortStars;
    public static boolean filter;

    public static final int PIN_NONE=0;
    public static final int PIN_SQUARE=1;
    public static final int PIN_CIRCLE_FIXED=2;
    public static final int PIN_CIRCLE_MOBILE=3;
    public static final int NB_PINMODE=4;

    public static int pinMode;

    public static final int BG_WHITE=0;
    public static final int BG_BLACK=1;
    public static final int BG_NAVY=2;
    public static final int NB_BACKGROUND=3;

    public static final int[][] RES={
	{320,240},
	{400,300},
	{512,384},
	{640,480},
	{800,600},
	{1024,768},
	{1280,1024},
	{1600,1200}};

    public static int nbResolutions, resolution;


    /**
     * The background mode
     */
    public static int backGround;

    /**
     * To avoid problems, it is better to have only one thread
     */
    //   public static Thread mainLoop;

    public static int screenWidth, screenHeight;
    public static int screenCenterX, screenCenterY;

    public static double scale1;
    public static double scale2;
    public static double scale3;

    /**
     * Used to memorize initial mouse click coordinate when turning
     * the point of view
     */
    public static int mouseX0,mouseY0;

    public static boolean doPaint;

    /** In sector mode, we view all the star in galactic coordinates */ 
    public static final int SECTOR_MODE=1;
    /** In system mode, we view a star and all the planets around it */
    public static final int SYSTEM_MODE=2;
    /** in planet mode, we view the surface of the planet */
    public static final int PLANET_MODE=3;
//        public static final int TOWN_MODE=4; // LATER

    /**
     * Mode tells wether we are looking at a sector, system, planet, ...
     */
    public static int mode;

    /** 
     * Are we turning point of view right now ?
     */
    public static boolean turning;

    public static MyFrame mainWindow;

    public static Insets mainInsets;

    /**
     * 3D Coordinates of the center of the view
     */
    public static double viewCenterX1,viewCenterY1,viewCenterZ1;
    public static double viewCenterX2,viewCenterY2,viewCenterZ2;

    public static int zoomLevel1,zoomLevel2,zoomLevel3;

    /**
     * Constructor
     */
    /*
    public SundogNavigator() {
    }
    */

    /**
     * Main
     * Here is to be put what does not work in constructor
     */
    public static void main(String[] args) {
	nbResolutions=RES.length;
	resolution=4;

	doPaint=false;

	mainWindow=new MyFrame();
	mainWindow.setTitle("Proto Sundog Navigator, v0.064");
	mainWindow.setResizable(false);

	mainWindow.show();

	resizeMainWindow();
	/*
	mainInsets=mainWindow.getInsets();
	mainWindow.setSize(X+mainInsets.left+mainInsets.right,Y+mainInsets.top+mainInsets.bottom);

	offDim=mainWindow.getSize();
	offImage=mainWindow.createImage(offDim.width,offDim.height);
	offGraphics=offImage.getGraphics();
	*/

	doPaint=true;

	turning=false;

	theta1=0.41;
	phi1=-0.55;
	thetaPhiToM1();
	theta2=Math.PI/4.0;
	phi2=-Math.PI/2.0;
	thetaPhiToM2();
	theta3=0.0;
	phi3=-Math.PI/8.0;
	thetaPhiToM3();

	viewCenterX1=0;
	viewCenterY1=0;
	viewCenterZ1=0;
	viewCenterX2=0;
	viewCenterY2=0;
	viewCenterZ2=0;

	mode=SECTOR_MODE;

	backGround=BG_BLACK;
	pinMode=PIN_CIRCLE_MOBILE;
	box=(pinMode==PIN_SQUARE);
	names=true;
	hideText=false;
	sortStars=true;
	filter=true;

	mainWindow.addMouseListener(mainWindow);
	mainWindow.addMouseMotionListener(mainWindow);
	mainWindow.addWindowListener(mainWindow);
	mainWindow.addKeyListener(mainWindow);

	//	mainLoop=new Thread(this, "Move");
	//	mainLoop.start();

	zoomLevel1=2;
	zoomLevel2=0;
	zoomLevel3=0;
	setScale1();
	setScale2();
	setScale3();

	File f=new File("DefaultSector.xml");
	currentSector=new Sector();
	currentSector.load(f);

	/*
	StarSystem s;
	currentSector=new Sector();
	s=new StarSystem();
	s.name=new String("Sol");
	s.x=10; s.y=10; s.z=10;
	currentSector.add(s);
	s=new StarSystem();
	s.name=new String("Alpha Centauri");
	s.x=5; s.y=0; s.z=0;
	currentSector.add(s);
	s=new StarSystem();
	s.name=new String("Tatooine");
	s.x=-8; s.y=6; s.z=2;
	currentSector.add(s);
	s=new StarSystem();
	s.name=new String("Jondd");
	s.x=1; s.y=-3; s.z=-8;
	currentSector.add(s);
	*/

	/*
	File ff=new File("Essai.xml");
	currentSector.save(ff);
	*/


	currentStarSystem=null;
	currentPlanet=null;

	mainWindow.repaint();
    }

    /**
     * Opens up a file dialog and loads the sector
     */
    public static void loadSector() {
	FileDialog fd=new FileDialog(mainWindow,"Choose Sector",FileDialog.LOAD);
	fd.show();
	String fileName=fd.getFile();
	if(fileName!=null) {
	    File f=new File(fileName);
	    Sector newSector=new Sector();
	    if(newSector.load(f)) {
		currentSector=newSector;
		currentStarSystem=null;
		currentPlanet=null;
		viewCenterX1=0;
		viewCenterY1=0;
		viewCenterZ1=0;
	    }
	}
    }

    /**
     * To call when corresponding zoomLevel has been changes
     */
    public static void setScale1() {
	scale1=((double)screenHeight)/2.0*scale10*Math.exp(0.3*(double)zoomLevel1);
    }
    public static void setScale2() {
	scale2=((double)screenHeight)/2.0*scale20*Math.exp(0.3*(double)zoomLevel2);
    }
    public static void setScale3() {
	scale3=((double)screenHeight)/2.0*scale30*Math.exp(0.3*(double)zoomLevel3);
    }

    /**
     * To call when resolution has been changed
     */
    public static void resizeMainWindow(){
	int X=RES[resolution][0];
	int Y=RES[resolution][1];

	mainInsets=mainWindow.getInsets();
	mainWindow.setSize(X+mainInsets.left+mainInsets.right,Y+mainInsets.top+mainInsets.bottom);

	offDim=mainWindow.getSize();
	offImage=mainWindow.createImage(offDim.width,offDim.height);
	offGraphics=offImage.getGraphics();

	screenWidth=X;
	screenHeight=Y;
	screenCenterX=X/2;
	screenCenterY=Y/2;
    }

    /**
     * Adds a pin z coordinate to a star system. Used when sorting the stars
     * according to their pin coordinates.
     */
    public static class StarWithZ {
	StarSystem s;
	double z;
    }

    /**
     * Comparator used when sorting the stars according to their pin z coordinates.
     */
    public static class StarComparator
	implements Comparator
    {
	public int compare(Object o1, Object o2) {
	    StarWithZ s1,s2;
	    s1=(StarWithZ) o1;
	    s2=(StarWithZ) o2;
	    return(s2.z-s1.z>0 ? 1 : s1==s2 ? 0 : -1);
	}
	// DO NOT override boolean equals(Object obj)
    }

    /**
     * The main window of the application. Almost everything happens here.
     */
    public static class MyFrame
	extends Frame
	implements MouseListener,
		   KeyListener,
		   MouseMotionListener,
		   WindowListener,
		   Runnable 
    {
	/**
	 * we redefine update so that it does not erase the background
	 */
	public void update(Graphics g) {
	    paint(g);
	}

	/**
	 * Paint (overriding).
	 * Does double buffering
	 */
	public void paint(Graphics g) {
	    if(doPaint) {
		Dimension d=getSize();
		if(d.width!=offDim.width || d.height!=offDim.height) {
		    offDim=getSize();
		    offImage=createImage(offDim.width,offDim.height);
		    offGraphics=offImage.getGraphics();
		}
		
		offPaint(offGraphics);
		
		g.drawImage(offImage,mainInsets.left,mainInsets.top,this); 
	    }
	}

	/** Consider this as the Paint method */
	public void offPaint(Graphics g) {
	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.WHITE); break;
	    case BG_BLACK : g.setColor(Color.BLACK); break;
	    case BG_NAVY  : g.setColor(new Color(0.4f,0.4f,1.0f)); break;
	    default : g.setColor(Color.WHITE);
	    }
	    g.fillRect(0,0,screenWidth,screenHeight);
	    
	    // To enhance readability, we call further subroutines
	    switch(mode) {
	    case SECTOR_MODE :
		drawSector(g);
		break;
	    case SYSTEM_MODE :
		drawStarSystem(g);
		break;
	    case PLANET_MODE :
		drawPlanet(g);
		break;
	    }
	}

	// -------------------------
	// Interfaces implementation
	// -------------------------

	// MouseListener
	
	public void mouseClicked(MouseEvent e) {
	}
	public void mouseEntered(MouseEvent e) {
	}
	public void mouseExited(MouseEvent e) {
	}
	public void mousePressed(MouseEvent e) {
	    if(e.getButton()==MouseEvent.BUTTON1) {
		Iterator i;
		int x,y,xx,yy;
		switch(mode) {
		case SECTOR_MODE :
		    i=currentSector.iterator();
		    StarSystem s;
		    xx=e.getX()-mainInsets.left;
		    yy=e.getY()-mainInsets.top;
		    currentStarSystem=null;
		    currentPlanet=null;
		    for(;i.hasNext();) {
			s=(StarSystem)(i.next());
			x=tox1(s.x,s.y,s.z);
			y=toy1(s.x,s.y,s.z);
			if((x-xx)*(x-xx)+(y-yy)*(y-yy)<=RAD*RAD) {
			    currentStarSystem=s;
			}
		    }
		    mainWindow.repaint();
		    break;
		case SYSTEM_MODE :
		    i=currentStarSystem.iterator();
		    Planet p;
		    double r,cos,sin;
		    xx=e.getX()-mainInsets.left;
		    yy=e.getY()-mainInsets.top;
		    currentPlanet=null;
		    for(;i.hasNext();) {
			p=(Planet)(i.next());
			r=AU*p.orbitRadius;
			cos=Math.cos(p.orbitInitAngle);
			sin=Math.sin(p.orbitInitAngle);
			x=tox2(r*cos,0,r*sin);
			y=toy2(r*cos,0,r*sin);
			if((x-xx)*(x-xx)+(y-yy)*(y-yy)<=RAD*RAD) {
			    currentPlanet=p;
			}
		    }
		    mainWindow.repaint();
		    break;
		case PLANET_MODE :
		    break;
		}
	    }
	    if((e.getButton() & MouseEvent.BUTTON2)!=0) {
		turning=true;
		switch(mode) {
		case SECTOR_MODE :
		    theta_0=theta1;
		    phi_0=phi1;
		    mouseX0=e.getX();
		    mouseY0=e.getY();
		    break;
		case SYSTEM_MODE :
		    theta_0=theta2;
		    phi_0=phi2;
		    mouseX0=e.getX();
		    mouseY0=e.getY();
		    break;
		case PLANET_MODE :
		    theta_0=theta3;
		    phi_0=phi3;
		    mouseX0=e.getX();
		    mouseY0=e.getY();
		    break;
		}
	    }
	}
	public void mouseReleased(MouseEvent e) {
	    turning=false;
	}
	
	// MouseListener

	public void mouseDragged(MouseEvent e) {
	    if(turning) {
		switch(mode) {
		case SECTOR_MODE :
		    theta1=theta_0-((double)(e.getX()-mouseX0))*angleFactor;
		    phi1=phi_0+((double)(mouseY0-e.getY()))*angleFactor;
		    thetaPhiToM1();
		    mainWindow.repaint();
		    break;
		case SYSTEM_MODE :
		    theta2=theta_0-((double)(e.getX()-mouseX0))*angleFactor;
		    phi2=phi_0+((double)(mouseY0-e.getY()))*angleFactor;
		    thetaPhiToM2();
		    mainWindow.repaint();
		    break;
		case PLANET_MODE :
		    theta3=theta_0-((double)(e.getX()-mouseX0))*angleFactor3;
		    phi3=phi_0+((double)(mouseY0-e.getY()))*angleFactor3;
		    thetaPhiToM3();
		    mainWindow.repaint();
		    break;
		}
	    }
	}
	public void mouseMoved(MouseEvent e) {
	}

	// KeyListener

	public void keyPressed(KeyEvent e) {
	}
	public void keyReleased(KeyEvent e) {
	}
	public void keyTyped(KeyEvent e) {	    
	    char c=e.getKeyChar();
	    if(c=='b' || c=='B') {
		backGround=(backGround+1) % NB_BACKGROUND;
		repaint();
	    }
	    if(c=='h' || c=='H') {
		hideText=!hideText;
		repaint();
	    }
	    if(c=='/' && resolution>0) {
		resolution--;
		resizeMainWindow();
		repaint();
	    }
	    if(c=='*' && resolution<nbResolutions-1) {
		resolution++;
		resizeMainWindow();
		repaint();
	    }
	    if(mode==SECTOR_MODE) {
		if(c=='+') {
		    zoomLevel1+=1;
		    setScale1();
		    repaint();
		}
		if(c=='-') {
		    zoomLevel1-=1;
		    setScale1();
		    repaint();
		}
		if((c=='c' || c=='C' || c=='5') && currentStarSystem!=null) {
		    viewCenterX1=currentStarSystem.x;
		    viewCenterY1=currentStarSystem.y;
		    viewCenterZ1=currentStarSystem.z;
		    repaint();
		}
		if(c=='f' || c=='F') {
		    filter=!filter;
		    repaint();
		}
		if(c=='g' || c=='G') {
		    pinMode=(pinMode+1) % NB_PINMODE;
		    box=(pinMode==PIN_SQUARE);
		    repaint();
		}
		if(c=='l' || c=='L') {
		    loadSector();
		    repaint();
		}
		if(c=='n' || c=='N') {
		    names=!names;
		    repaint();
		}
		if(c=='z' || c=='Z') {
		    if(currentStarSystem!=null) {
			mode=SYSTEM_MODE;
//			zoomLevel2=0; // Is it better not to remeber the zoom level in that mode ?
//			setScale2();
			viewCenterX2=0;
			viewCenterY2=0;
			viewCenterZ2=0;
			repaint();
		    }
		}
	    }
	    else if(mode==SYSTEM_MODE) {
		if(c=='x' || c=='X') {
		    mode=SECTOR_MODE;
		    repaint();
		}
		if(c=='z' || c=='Z') {
		    if(currentPlanet!=null) {
			mode=PLANET_MODE;
			repaint();
		    }
		}
		if(c=='+') {
		    zoomLevel2+=1;
		    setScale2();
		    repaint();
		}
		if(c=='-') {
		    zoomLevel2-=1;
		    setScale2();
		    repaint();
		}
	    }
	    else if(mode==PLANET_MODE) {
		if(c=='x' || c=='X') {
		    mode=SYSTEM_MODE;
		    repaint();
		}
	    }
	}

	// WindowListener

	public void windowActivated(WindowEvent e) {
	}
	public void windowClosed(WindowEvent e) {
	}
	public void windowClosing(WindowEvent e) {
	    //	mainLoop=null;
	    /*	mainWindow.setVisible(false);
		mainWindow=null;*/
	    System.exit(0);
	}
	public void windowDeactivated(WindowEvent e) {
	}
	public void windowDeiconified(WindowEvent e) {
	}
	public void windowIconified(WindowEvent e) {
	}
	public void windowOpened(WindowEvent e) {
	}
 

	// -----------
	// Threads
	// -----------

	// it would be better to have only one supplementary thread
	// ... this remains to be done

	public void run() {
	}
    }

    /**
     * Converts a screen intermediate coordinate into a pixel coordinate
     * in the main window
     */
    public static int xtoi1(double x) {
	return screenCenterX+(int)(x*scale1);
    }
    public static int ytoi1(double y) {
	return screenCenterY-(int)(y*scale1);
    }
    public static int xtoi2(double x) {
	return screenCenterX+(int)(x*scale2);
    }
    public static int ytoi2(double y) {
	return screenCenterY-(int)(y*scale2);
    }
    public static int xtoi3(double x) {
	return screenCenterX+(int)(x*scale3);
    }
    public static int ytoi3(double y) {
	return screenCenterY-(int)(y*scale3);
    }

    /**
     * converts x,y,z into x,y via the matrix v1
     */
    public static int tox1(double x, double y, double z) {
	return xtoi1((x-viewCenterX1)*v1M11+(y-viewCenterY1)*v1M12+(z-viewCenterZ1)*v1M13);
    }
    public static int toy1(double x, double y, double z) {
	return ytoi1((x-viewCenterX1)*v1M21+(y-viewCenterY1)*v1M22+(z-viewCenterZ1)*v1M23);
    }

    /**
     * converts x,y,z into x,y via the matrix v2
     */
    public static int tox2(double x, double y, double z) {
	return xtoi2((x-viewCenterX2)*v2M11+(y-viewCenterY2)*v2M12+(z-viewCenterZ2)*v2M13);
    }
    public static int toy2(double x, double y, double z) {
	return ytoi2((x-viewCenterX2)*v2M21+(y-viewCenterY2)*v2M22+(z-viewCenterZ2)*v2M23);
    }

    /**
     * converts x,y,z into x,y via the matrix v3
     */
    public static int tox3(double x, double y, double z) {
	return xtoi3(x*v3M11+y*v3M12+z*v3M13);
    }
    public static int toy3(double x, double y, double z) {
	return ytoi3(x*v3M21+y*v3M22+z*v3M23);
    }
    public static double toz3(double x, double y, double z) {
	return (x*v3M31+y*v3M32+z*v3M33);
    }

    /**
     * Special
     */
    public static double toZv(double x, double y, double z) {
	return ((x-viewCenterX1)*v1M31+0+(z-viewCenterZ1)*v1M33);
    }

    /*
    public static Color starColor(StarSystem s) {
	int c=s.getColor();
	Color col;
	switch(c) {
	case StarSystem.WHITE :         col=new Color(0.0f,0.0f,1.0f); break;
	case StarSystem.BRIGHT_BLUE :   col=new Color(0.0f,0.5f,1.0f); break;
	case StarSystem.BLUE :          col=new Color(0.0f,0.9f,0.9f); break;
	case StarSystem.BRIGHT_YELLOW : col=new Color(0.9f,0.9f,0.9f); break;
	case StarSystem.YELLOW :        col=Color.YELLOW; break;
	case StarSystem.ORANGE :        col=Color.ORANGE; break;
	case StarSystem.RED :           col=Color.RED; break;
	case StarSystem.DARK_RED :      col=new Color(0.7f,0,0); break;
	default : col=Color.ORANGE;
	}
	return(col);	
    }
    */

    /**
     * This draws the star s in g (for mode SECTOR_MODE)
     */
    public static void drawStar(Graphics g, StarSystem s) {
	int x=tox1(s.x,s.y,s.z);
	int y=toy1(s.x,s.y,s.z);

	// Draw a vertical line from the ztar to the plane y=0

	if(pinMode!=PIN_NONE) {
	    int x1,y1;
	    if(pinMode==PIN_CIRCLE_MOBILE) {
		x1=tox1(s.x,viewCenterY1,s.z);
		y1=toy1(s.x,viewCenterY1,s.z);
	    }
	    else {
		x1=tox1(s.x,0*M,s.z);
		y1=toy1(s.x,0*M,s.z);
	    }
	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.lightGray); break;
	    case BG_BLACK : g.setColor(Color.gray); break;
	    case BG_NAVY  : g.setColor(Color.BLACK); break;
	    default : g.setColor(Color.lightGray);
	    }
	    g.drawLine(x,y,x1,y1);
	}

	// Draw the star itself

	g.setColor(s.mainStar.getColor());
	int r=(int) (Math.max(2.0,Math.min(8.0,4.0-s.mainStar.magnitude*0.4)));
	g.fillOval(x-r,y-r,2*r,2*r);

	// Draw it's name

	if(names) {
	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.BLACK); break;
	    case BG_BLACK : g.setColor(Color.WHITE); break;
	    case BG_NAVY  : g.setColor(Color.WHITE); break;
	    default : g.setColor(Color.BLACK);
	    }
	    g.drawString(s.name,x,y-(r+7));
	}

	// If selected, draw selection mark

	if(s==currentStarSystem) {
	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.BLUE); break;
	    case BG_BLACK : g.setColor(Color.GREEN); break;
	    case BG_NAVY  : g.setColor(Color.RED); break;
	    default : g.setColor(Color.BLUE);
	    }
	    g.drawOval(x-(r+2),y-(r+2),2*(r+2),2*(r+2));
	}
    }

    /**
     * This draws all the stars, and decorations according to the
     * options (for mode SECTOR_MODE)
     */
    public static void drawSector(Graphics g) {
	if(currentSector!=null) {
	    Iterator i;
	    StarSystem s;
	    int n;

	    int count;
	    i=currentSector.iterator();
	    for(count=0;i.hasNext();) {
		s=(StarSystem)(i.next());
		count++;
	    }

	    StarWithZ[] swz=new StarWithZ[count];
	    i=currentSector.iterator();
	    for(n=0;i.hasNext();) {
		s=(StarSystem)(i.next());
		swz[n]=new StarWithZ();
		swz[n].s=s;
		swz[n].z=toZv(s.x,s.y,s.z);
		n++;
	    }

	    Arrays.sort(swz,new StarComparator());

	    if(currentSector.comment!=null) {
		g.setColor(Color.RED);
		g.drawString(currentSector.comment,screenWidth/2,30);
	    }

	    if(currentStarSystem!=null) {
		if(currentStarSystem.comment!=null) {
		    g.setColor(Color.RED);
		    g.drawString(currentStarSystem.comment,10,screenHeight-50);
		}
	    }

	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.gray); break;
	    case BG_BLACK : g.setColor(Color.lightGray); break;
	    case BG_NAVY  : g.setColor(Color.BLACK); break;
	    default : g.setColor(Color.gray);
	    }
	    int h=30;
	    g.drawString(String.valueOf(count)+" Star Systems",30,h);
	    h+=30;
	    g.drawString("Selected : "+(currentStarSystem==null ? "none" : currentStarSystem.name),30,h);
	    h+=20;
	    if(!hideText) {
		if(currentStarSystem!=null) {
		    g.drawString("           x="+
			     String.valueOf(currentStarSystem.x)+
			     " y="+String.valueOf(currentStarSystem.y)+
			     " z="+String.valueOf(currentStarSystem.z)
			     ,30,h);
		}
		h+=30;
		g.drawString("Press 'g' to toggle grid",30,h);
		h+=20;
		g.drawString("Press 'n' to toggle names",30,h);
		h+=20;
		g.drawString("Press 'b' to toggle background",30,h);
		h+=20;
		g.drawString("Press 'c' or '5' to center view on selected star",30,h);
		h+=20;
		g.drawString("Press 'f' to filter out far stars",30,h);
		h+=20;
		g.drawString("Press +/- to zoom in/out",30,h);
		h+=20;
		g.drawString("Press / or * to decrease or increase window size",30,h);
		h+=20;
		g.drawString("Press 'z' to see selected system",30,h);
		h+=20;
		g.drawString("Press 'h' hide this text",30,h);
		h+=30;
		g.drawString("Press 'l' to load a new starmap",30,h);
		//	    g.drawString("theta "+String.valueOf(theta1),30,50);
		//	    g.drawString("phi
		//	    "+String.valueOf(phi1),30,70);
	    }

	    int x1,y1,x2,y2;
	    switch(backGround) {
	    case BG_WHITE : g.setColor(new Color(1.0f,0.7f,0.7f)); break; // light red
	    case BG_BLACK : g.setColor(new Color(0.7f,0.0f,0.0f)); break; // dark red
	    case BG_NAVY  : g.setColor(new Color(1.0f,0.7f,0.7f)); break; 
	    default :       g.setColor(new Color(1.0f,0.7f,0.7f));
	    }
	    if(pinMode==PIN_SQUARE) {
		final int NB=10;
		double l;
		for(n=0; n<=NB; n++) {
		    l=1.0-2.0*((double) n)/((double) NB);
		    x1=tox1(l*M,0,M);
		    y1=toy1(l*M,0,M);
		    x2=tox1(l*M,0,-M);
		    y2=toy1(l*M,0,-M);
		    g.drawLine(x1,y1,x2,y2);
		    x1=tox1(M,0,l*M);
		    y1=toy1(M,0,l*M);
		    x2=tox1(-M,0,l*M);
		    y2=toy1(-M,0,l*M);
		    g.drawLine(x1,y1,x2,y2);
		}
	    }
	    if(pinMode==PIN_CIRCLE_FIXED) {
		final int NBR=24,NBC=10,NBA=60;
		int k;
		double cos,sin,alpha,cos_,sin_,ca,sa,r;
		for(n=0; n<NBR; n++) {
		    alpha=2.0*Math.PI*((double)n)/((double)NBR);
		    cos=Math.cos(alpha);
		    sin=Math.sin(alpha);
		    x1=tox1(0,0,0);
		    y1=toy1(0,0,0);
		    x2=tox1(cos*M,0,sin*M);
		    y2=toy1(cos*M,0,sin*M);
		    g.drawLine(x1,y1,x2,y2);
		}
		cos_=1;
		sin_=0;
		alpha=2.0*Math.PI/(double)NBA;
		ca=Math.cos(alpha);
		sa=Math.sin(alpha);
		for(k=0; k<NBA; k++) {
		    //		    alpha=2.0*Math.PI*((double)(k+1))/(double)NBA;
		    cos=ca*cos_-sa*sin_;
		    sin=sa*cos_+ca*sin_;
		    for(n=1; n<=NBC; n++) {
			r=((double)n)/((double)NBC)*M;
			x1=tox1(r*cos_,0,r*sin_);
			y1=toy1(r*cos_,0,r*sin_);
			x2=tox1(r*cos,0,r*sin);
			y2=toy1(r*cos,0,r*sin);
			g.drawLine(x1,y1,x2,y2);
		    }
		    cos_=cos;
		    sin_=sin;
		}
	    }
	    if(pinMode==PIN_CIRCLE_MOBILE) {
		//
		// This one makes too much computations compared to
		// what is required !
		// To be optimized, one of these days...
		//
		final int NBR=24,NBC=10,NBA=60;
		int k;
		double cos,sin,alpha,cos_,sin_,ca,sa,r;
		double lM=((double) screenHeight)/scale1/2.0;
		for(n=0; n<NBR; n++) {
		    alpha=2.0*Math.PI*((double)n)/((double)NBR);
		    cos=Math.cos(alpha);
		    sin=Math.sin(alpha);
		    x1=tox1(viewCenterX1,viewCenterY1,viewCenterZ1);
		    y1=toy1(viewCenterX1,viewCenterY1,viewCenterZ1);
		    x2=tox1(cos*lM+viewCenterX1,viewCenterY1,sin*lM+viewCenterZ1);
		    y2=toy1(cos*lM+viewCenterX1,viewCenterY1,sin*lM+viewCenterZ1);
		    g.drawLine(x1,y1,x2,y2);
		}
		cos_=1;
		sin_=0;
		alpha=2.0*Math.PI/(double)NBA;
		ca=Math.cos(alpha);
		sa=Math.sin(alpha);
		for(k=0; k<NBA; k++) {
		    //		    alpha=2.0*Math.PI*((double)(k+1))/(double)NBA;
		    cos=ca*cos_-sa*sin_;
		    sin=sa*cos_+ca*sin_;
		    for(n=1; n<=NBC; n++) {
			r=((double)n)/((double)NBC)*lM;
			x1=tox1(r*cos_+viewCenterX1,viewCenterY1,r*sin_+viewCenterZ1);
			y1=toy1(r*cos_+viewCenterX1,viewCenterY1,r*sin_+viewCenterZ1);
			x2=tox1(r*cos+viewCenterX1,viewCenterY1,r*sin+viewCenterZ1);
			y2=toy1(r*cos+viewCenterX1,viewCenterY1,r*sin+viewCenterZ1);
			g.drawLine(x1,y1,x2,y2);
		    }
		    cos_=cos;
		    sin_=sin;
		}
	    }

	    if(!filter) {
		for(n=0; n<count; n++) {
		    drawStar(g,swz[n].s);
		}
	    }
	    else {
		for(n=0; n<count; n++) {
		    if(((swz[n].s.x-viewCenterX1)*(swz[n].s.x-viewCenterX1)+
			(swz[n].s.y-viewCenterY1)*(swz[n].s.y-viewCenterY1)+
			(swz[n].s.z-viewCenterZ1)*(swz[n].s.z-viewCenterZ1))
		       *scale1*scale1
		       < (screenHeight/2)*(screenHeight/2))
		    drawStar(g,swz[n].s);
		}
	    }

	    if(box) {
		g.setColor(Color.RED);
		int x3,y3,x4,y4,x5,y5,x6,y6,x7,y7,x8,y8;
		x1=tox1(M,H,M);
		y1=toy1(M,H,M);
		x2=tox1(-M,H,M);
		y2=toy1(-M,H,M);
		x3=tox1(-M,-H,M);
		y3=toy1(-M,-H,M);
		x4=tox1(M,-H,M);
		y4=toy1(M,-H,M);
		x5=tox1(M,H,-M);
		y5=toy1(M,H,-M);
		x6=tox1(-M,H,-M);
		y6=toy1(-M,H,-M);
		x7=tox1(-M,-H,-M);
		y7=toy1(-M,-H,-M);
		x8=tox1(M,-H,-M);
		y8=toy1(M,-H,-M);
		g.drawLine(x1,y1,x2,y2);
		g.drawLine(x2,y2,x3,y3);
		g.drawLine(x3,y3,x4,y4);
		g.drawLine(x4,y4,x1,y1);
		g.drawLine(x1,y1,x5,y5);
		g.drawLine(x2,y2,x6,y6);
		g.drawLine(x3,y3,x7,y7);
		g.drawLine(x4,y4,x8,y8);
		g.drawLine(x5,y5,x6,y6);
		g.drawLine(x6,y6,x7,y7);
		g.drawLine(x7,y7,x8,y8);
		g.drawLine(x8,y8,x5,y5);
	    }
	}
	else{
	    g.drawString("No Sector !",30,30);
	}
	switch(backGround) {
	case BG_WHITE : g.setColor(Color.BLACK); break; // light red
	case BG_BLACK : g.setColor(Color.WHITE); break; // dark red
	case BG_NAVY  : g.setColor(Color.BLACK); break; 
	default :       g.setColor(Color.BLACK);
	}
	g.drawRect(0,0,screenWidth-1,screenHeight-1);

	if(filter) {
	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.BLACK); break; // light red
	    case BG_BLACK : g.setColor(Color.WHITE); break; // dark red
	    case BG_NAVY  : g.setColor(Color.WHITE); break; 
	    default :       g.setColor(Color.WHITE);
	    }
	    g.drawOval(screenCenterX-screenHeight/2,screenCenterY-screenHeight/2,screenHeight,screenHeight);
	    // height has to be less thant width !!!
	}
    }

    /**
     * Set color according to background and context
     * To be implemented
     */
    public static void setColor(Graphics g, int context) {
    }

    /**
     * This draws the central star and the orbit of planets,
     * options (for mode SYSTEM_MODE)
     */
    public static void drawStarSystem(Graphics g) {

	// Draw comment

	if(currentStarSystem.comment!=null) {
	    g.setColor(Color.RED);
	    g.drawString(currentStarSystem.comment,screenWidth/2,30);
	}


	// Draw the text

	if(!hideText) {
	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.gray); break;
	    case BG_BLACK : g.setColor(Color.lightGray); break;
	    case BG_NAVY  : g.setColor(Color.BLACK); break;
	    default : g.setColor(Color.gray);
	    }
	    int h=30;
	    g.drawString(currentStarSystem.name,30,h);
	    h+=30;
	    g.drawString("Selected : "+(currentPlanet==null ? "none" : currentPlanet.name),30,h);
	    h+=30;
	    //	    g.drawString("Press 'g' to toggle grid",30,h);
	    h+=20;
	    //	    g.drawString("Press 'n' to toggle names",30,h);
	    h+=20;
	    //	    g.drawString("Press 'c' to center view on selected object",30,h);
	    h+=20;
	    g.drawString("Press 'b' to toggle background",30,h);
	    h+=20;
	    g.drawString("Press +/- to zoom in/out",30,h);
	    h+=20;
	    g.drawString("Press / or * to decrease or increase window size",30,h);
	    h+=20;
	    g.drawString("Press 'z' to see selected planet",30,h);
	    h+=20;
	    g.drawString("Press 'x' to return to starmap",30,h);
	    h+=20;
	    g.drawString("Press 'h' hide this text",30,h);
	    //	    g.drawString("theta "+String.valueOf(theta1),30,50);
	    //	    g.drawString("phi
	    //	    "+String.valueOf(phi1),30,70);
	}

	// Draw the central star

	g.setColor(currentStarSystem.mainStar.getColor());
	int diam=Math.max((int)(currentStarSystem.mainStar.diameter*scale2),8);
	g.fillOval(screenCenterX-diam/2,screenCenterY-diam/2,diam,diam);

	// Draw the orbits of planets

	final int NBA=60;
	int k;
	double cos,sin,alpha,cos_,sin_,ca,sa,r;
	int x1,y1,x2,y2;
	Iterator i;
	Planet p;
	for(i=currentStarSystem.iterator();i.hasNext();) {

	    // orbit

	    g.setColor(new Color(0.7f,0,0));
	    p=(Planet) i.next();
	    r=AU*p.orbitRadius;
	    cos_=1;
	    sin_=0;
	    alpha=2.0*Math.PI/(double)NBA;
	    ca=Math.cos(alpha);
	    sa=Math.sin(alpha);
	    for(k=0;k<NBA;k++) {
		cos=ca*cos_-sa*sin_;
		sin=sa*cos_+ca*sin_;
		x1=tox2(r*cos_,0,r*sin_);
		y1=toy2(r*cos_,0,r*sin_);
		x2=tox2(r*cos,0,r*sin);
		y2=toy2(r*cos,0,r*sin);
		g.drawLine(x1,y1,x2,y2);
		cos_=cos;
		sin_=sin;
	    }

	    // planet

	    g.setColor(new Color(0,1.0f,0));
	    cos=Math.cos(p.orbitInitAngle);
	    sin=Math.sin(p.orbitInitAngle);
	    x1=tox2(r*cos,0,r*sin);
	    y1=toy2(r*cos,0,r*sin);
	    g.fillOval(x1-RAD,y1-RAD,2*RAD,2*RAD);

	    // if selected, draw selection mark

	    if(p==currentPlanet) {
		switch(backGround) {
		case BG_WHITE : g.setColor(Color.BLUE); break;
		case BG_BLACK : g.setColor(Color.GREEN); break;
		case BG_NAVY  : g.setColor(Color.RED); break;
		default : g.setColor(Color.BLUE);
		}
		g.drawOval(x1-(RAD+2),y1-(RAD+2),2*(RAD+2),2*(RAD+2));
	    }

	    // planet name

	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.BLACK); break;
	    case BG_BLACK : g.setColor(Color.WHITE); break;
	    case BG_NAVY  : g.setColor(Color.WHITE); break;
	    default : g.setColor(Color.BLACK);
	    }
	    g.drawString(p.name,x1,y1-(RAD+7));
	}
    }

    /**
     * To be implemented. This would be great to have a 3D planet.
     */
    public static void drawPlanet(Graphics g) {

	// Draw the text

	if(!hideText) {
	    switch(backGround) {
	    case BG_WHITE : g.setColor(Color.gray); break;
	    case BG_BLACK : g.setColor(Color.lightGray); break;
	    case BG_NAVY  : g.setColor(Color.BLACK); break;
	    default : g.setColor(Color.gray);
	    }
	    int h=30;
	    g.drawString(currentPlanet.name,30,h);
	    h+=30;
	    //	    g.drawString("Selected : "+(currentPlanet==null ? "none" : currentPlanet.name),30,h);
	    //	    h+=30;
	    //	    g.drawString("Press 'g' to toggle grid",30,h);
	    //	    h+=20;
	    //	    g.drawString("Press 'n' to toggle names",30,h);
	    //	    h+=20;
	    h+=20;
	    g.drawString("Press 'b' to toggle background",30,h);
	    h+=20;
	    //	    g.drawString("Press +/- to zoom in/out",30,h);
	    //	    h+=20;
	    g.drawString("Press / or * to decrease or increase window size",30,h);
	    h+=20;
	    //	    g.drawString("Press 'z' to see selected planet",30,h);
	    //	    h+=20;
	    g.drawString("Press 'x' to return to system map",30,h);
	    h+=20;
	    g.drawString("Press 'h' hide this text",30,h);
	    //	    g.drawString("theta "+String.valueOf(theta1),30,50);
	    //	    g.drawString("phi
	    //	    "+String.valueOf(phi1),30,70);
	}

	//

	final int CR=4;
	double r=((double) screenHeight)/scale3/4.0;
	int x1,y1,x2,y2;

	// draw back cities

	Iterator i;
	i=currentPlanet.iterator();
	City c;
	double x,y,z,la,lo;
	for(;i.hasNext();) {
	    c=(City)(i.next());
	    la=c.latitude/180.0*Math.PI;
	    lo=c.longitude/180.0*Math.PI;
	    z=r*Math.cos(lo)*Math.cos(la);
	    x=r*Math.sin(lo)*Math.cos(la);
	    y=r*Math.sin(la);
	    if(toz3(x,y,z)>0) {
		x1=tox3(x,y,z);
		y1=toy3(x,y,z);
		g.setColor(c.type==City.STARPORT ? Color.white : Color.gray);
		g.fillOval(x1-CR,y1-CR,2*CR,2*CR);
	    }; 
	}

	// Draw planet's boundary

	g.setColor(Color.green);
	g.drawOval(screenCenterX-screenHeight/4,screenCenterY-screenHeight/4,
		   screenHeight/2,screenHeight/2);

	// Draw meridians

	final int NBM=24,NBAM=30;
	int j,k;
	double cos,sin,alpha,cos_,sin_,ca,sa,sb,cb;
	for(j=0; j<NBM; j++) {
	    cos_=1;
	    sin_=0;
	    alpha=Math.PI/(double)NBAM;
	    ca=Math.cos(alpha);
	    sa=Math.sin(alpha);
	    alpha=j*2.0*Math.PI/(double)NBM;
	    cb=Math.cos(alpha);
	    sb=Math.sin(alpha);
	    for(k=0; k<NBAM; k++) {
		cos=ca*cos_-sa*sin_;
		sin=sa*cos_+ca*sin_;
		x1=tox3(r*cb*sin_,r*cos_,r*sb*sin_);
		y1=toy3(r*cb*sin_,r*cos_,r*sb*sin_);
		x2=tox3(r*cb*sin,r*cos,r*sb*sin);
		y2=toy3(r*cb*sin,r*cos,r*sb*sin);
		if(toz3(r*cb*sin,r*cos,r*sb*sin)<0) {
		    g.drawLine(x1,y1,x2,y2);
		};
		cos_=cos;
		sin_=sin;
	    }
	}

	// Draw Parallels

	final int NBP=12,NBAP=60;
	for(j=1; j<NBP; j++) {
	    cos_=1;
	    sin_=0;
	    alpha=2.0*Math.PI/(double)NBAP;
	    ca=Math.cos(alpha);
	    sa=Math.sin(alpha);
	    alpha=j*Math.PI/(double)NBP;
	    cb=Math.cos(alpha);
	    sb=Math.sin(alpha);
	    for(k=0; k<NBAP; k++) {
		cos=ca*cos_-sa*sin_;
		sin=sa*cos_+ca*sin_;
		x1=tox3(r*cos_*sb,r*cb,r*sin_*sb);
		y1=toy3(r*cos_*sb,r*cb,r*sin_*sb);
		x2=tox3(r*cos*sb,r*cb,r*sin*sb);
		y2=toy3(r*cos*sb,r*cb,r*sin*sb);
		if(toz3(r*cos*sb,r*cb,r*sin*sb)<0) {
		    g.drawLine(x1,y1,x2,y2);
		};
		cos_=cos;
		sin_=sin;
	    }
	}

	// draw front cities

	i=currentPlanet.iterator();
	for(;i.hasNext();) {
	    c=(City)(i.next());
	    la=c.latitude/180.0*Math.PI;
	    lo=c.longitude/180.0*Math.PI;
	    z=r*Math.cos(lo)*Math.cos(la);
	    x=r*Math.sin(lo)*Math.cos(la);
	    y=r*Math.sin(la);
	    if(toz3(x,y,z)<0) {
		x1=tox3(x,y,z);
		y1=toy3(x,y,z);
		g.setColor(c.type==City.STARPORT ? Color.red : Color.magenta);
		g.fillOval(x1-CR,y1-CR,2*CR,2*CR);
		g.drawString(c.name,x1,y1-(CR+7));
	    }; 
	}
    }

    /**
     * This creates the matrix to transform 3D galactic coordinates
     * into 2D intermediate screen coordinates (for final ones, use
     * xtoi and ytoy). The third line is there to compute which star
     * is closer to the point of view.
     */
    public static void thetaPhiToM1() {
	double c,cc,s,ss;
	c=Math.cos(theta1);
	s=Math.sin(theta1);
	cc=Math.cos(phi1);
	ss=Math.sin(phi1);
	v1M11=c;     v1M12=0;  v1M13=s;
	v1M21=s*ss;  v1M22=cc; v1M23=-c*ss;
	v1M31=-s*cc; v1M32=ss; v1M33=c*cc;
    }

    public static void thetaPhiToM2() {
	double c,cc,s,ss;
	c=Math.cos(theta2);
	s=Math.sin(theta2);
	cc=Math.cos(phi2);
	ss=Math.sin(phi2);
	v2M11=c; v2M12=0; v2M13=s;
	v2M21=s*ss; v2M22=cc; v2M23=-c*ss;
    }

    public static void thetaPhiToM3() {
	double c,cc,s,ss;
	c=Math.cos(theta3);
	s=Math.sin(theta3);
	cc=Math.cos(phi3);
	ss=Math.sin(phi3);
	v3M11=c; v3M12=0; v3M13=s;
	v3M21=s*ss; v3M22=cc; v3M23=-c*ss;
	v3M31=-s*cc; v3M32=ss; v3M33=c*cc;
    }

}
